// Random numbers 

#include "types.h"
#include "large_math.h"
#include "hash_func.h"
#include "xtea.h"
#include "idea.h"
#include "random.h"
#include "system_random.h"
#include <time.h>
#include "serialize.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <tchar.h>
#include "system_critical_section.h"

//#include <wchar.h>
//#include <tchar_s.h>
//#include <stdio_s.h>



static SYSTEM_CRITICAL_SECTION main_rng_critical_section; // Must be static

static inline void mix_round(u32 *seed, u32 x);


// The main seed is 1 KB size

static const u32 seed_size_u32 = 256;

static u32 default_seed[seed_size_u32]; // main seed
static u32 cheap_seed[2]; // To generate junk


//--------------------------------------
// Initialize seed

#ifndef _UNICODE

static char seed_path[MAX_PATH + 1] = "seed.bin";

void initialize_seed(const char *path)
{
if(path != NULL)
 {
 strncpy(seed_path, path, sizeof(seed_path));
 }

u8 buff[seed_size_u32*sizeof(u32)];

FILE *file = fopen(seed_path,"rb");

if(file != NULL)
 {
 fread(buff, sizeof(buff), 1, file); 
 fclose(file);
 
 collect_entropy(buff, sizeof(buff));
 }

// Wash seed

collect_entropy();

file = fopen(seed_path,"wb");
fwrite(default_seed, seed_size_u32*sizeof(u32), 1, file);
fclose(file);

// Initialize cheap seed
 
time_t t;
time(&t);
Hash_Block_X_64(cheap_seed, (u8 *)&t, sizeof(time_t));

// Cleanup

cheap_random_number(buff, seed_size_u32); 
reset_stack();
}
#else

static TCHAR seed_path[MAX_PATH + 1] = _T("seed.bin");

void initialize_seed(const TCHAR *path)
{
if(path != NULL)
 {
 _tcsncpy(seed_path, path, MAX_PATH + 1);
 }

u8 buff[seed_size_u32*sizeof(u32)];

FILE *file;

if(!_tfopen_s(&file,seed_path,_T("rb")))
 {
 fread(buff, sizeof(buff), 1, file); 
 fclose(file);
 
 collect_entropy(buff, sizeof(buff));
 }

// Wash seed

collect_entropy();

if(!_tfopen_s(&file,seed_path,_T("wb")))
 {
 fwrite(default_seed, seed_size_u32*sizeof(u32), 1, file);
 fclose(file);
 }

// Initialize cheap seed
 
time_t t;
time(&t);
Hash_Block_X_64(cheap_seed, (u8 *)&t, sizeof(time_t));

// Cleanup

cheap_random_number(buff, seed_size_u32); 
reset_stack();
}

#endif

//--------------------------
// PRNG machine 
//
static u32 generan(u32 *state, u32 x)
{

u32 tmp = state[0] ^ x;
x += k_mix(state[0]);

for(u32 i = 1; i < seed_size_u32; i++)
 {
 state[i-1] = state[i] ^ x;
 x += k_mix(state[i]);
 }

state[seed_size_u32 - 1] = tmp;

return x;
}
//-----------------
// Secure PRNG machine
//
static u32 secure_generan(u32 *state)
{
u32 xx[6]; 


u32 x = lcg64(state);

xx[0] = x ^ state[2]; // feedback
x += k_mix(state[2]);
 
for(u32 i = 3; i < seed_size_u32; i++)
 {
 state[i-1] = state[i] ^ x;
 x += k_mix(state[i]);
 } // i
 
 
xx[1] = x;
xx[2] = state[seed_size_u32 - 1 - 83];
xx[3] = state[seed_size_u32 - 1 - 167];
 
 // Input: 
 // 0: feedback
 // 1: generated x
 // 2: state[172] coprime taps separated by about 1/3 of length
 // 3: state[88]
 //------
 // Output:
 // 0: feedback ( -> state[255])
 // 1: output value

// Hash function
 
xx[4] = xx[0];
xx[5] = xx[1];

XTEA_Encode(xx, xx + 2);
 
state[seed_size_u32 - 1] = xx[0]; // feedback

return xx[1]; // output
}


//----------------------------------
// Mix entropy into main seed
//

// System provided

void collect_entropy()
{
const u32 buff_size_bytes = 1024; // Draw 1K

u8 buff[buff_size_bytes];    

Get_System_Provided_Entropy(buff, buff_size_bytes);

collect_entropy(buff, buff_size_bytes);

time_t t;

time(&t);

collect_entropy(&t, sizeof(time_t));
}

//----
// u32[2] for legacy compatibility
//
void collect_entropy(u32 const* data)
{
collect_entropy(data, 2*sizeof(u32));
}



void collect_entropy(void const *data, const u32 data_size_bytes)
{

if((data == NULL) || (data_size_bytes == 0))
 {
 collect_entropy();
 return;
 }


u32 lock[seed_size_u32];
u32 key[4];

zero(key, 4);

main_rng_critical_section.EnterCriticalSection();

copy(lock, default_seed, seed_size_u32);

main_rng_critical_section.LeaveCriticalSection();


// Smear entropy (fast, unsecure)

u32 cnt = 0;

while(cnt < 4) // Make no less then 4 iterations to shuffle and initialize key
 {
 u8 const * data_ptr = (u8 const *)data;
 u32 size = data_size_bytes;   

 while(size)
  {
  u32 x;  
  size = fetch_dwords(&x, 1, data_ptr, size);
  data_ptr += sizeof(u32);  

  // Absorb

  x = generan(lock, x + cnt);
  
  // Collect key
 
  key[(cnt & 3)] ^= x;
  cnt++;
  } // size
 } // cnt

// Encipher (8 rounds, doesn't have to be very strong)

XXTEA_Encode(lock, seed_size_u32, key, 8);

// Lock up

main_rng_critical_section.EnterCriticalSection();

for(u32 i = 0; i < seed_size_u32; i++)
 {
 default_seed[i] ^= lock[i];
 lock[i] = 0;
 }

main_rng_critical_section.LeaveCriticalSection();


zero(key, 4);
reset_stack();
}

void reset_stack(void)
{
const u32 dummy_size = 256;    
u32 dummy[dummy_size];
cheap_random_number(dummy, dummy_size);
}

//----------------------------------------------------
// No need to protect cheap rng with critical sections

u32 cheap_random_number()
{
return lcg64(cheap_seed);
}
 
void cheap_random_number(u32 *x, const u32 length_x32)
{
for(u32 i = 0; i < length_x32; i++) x[i] = lcg64(cheap_seed);
}

void cheap_random_number(u8 *x, const u32 size_of_x_bytes)
{
for(u32 i = 0; i < size_of_x_bytes; i++) x[i] = (u8)lcg64(cheap_seed);
}

//---------------------------
// Simple 64-bit LCG
// Do NOT change since it is used in keystream generation and hash functions
//
u32 lcg64(u32 *state)
{
u64 x = state[0] | (((u64)state[1])<<32);
x = (x << 5) + x + 0x71E86A1CD2D0E1DBllu;
state[0] = (u32)x;
state[1] = (u32)(x>>32);
return (state[0] ^ state[1]);
}


//--------------------------------------------
// Generation of strong random numbers
//

void random_number(u32 *x, const u32 length_x32)
{
main_rng_critical_section.EnterCriticalSection();

// Prehash
collect_entropy();

for(u32 i = 0; i < length_x32; i++) x[i] = secure_generan(default_seed);

// Posthash

collect_entropy();

main_rng_critical_section.LeaveCriticalSection();

reset_stack();
}

//----------------------
//  x ^= C
//  x = ((x << 32) - x);
// Full 128 bit period
//
u32 nlcg128(u32 *state)
{
const u32 C = 0xAA551EE1; // Any odd number

u32 tmp32 = state[0];
tmp32 ^= C;
s64 tmp64 = 0;
tmp64 -= tmp32;
state[0] = (u32)tmp64;
tmp64 >>= 32; // Arithmetic shift
tmp64 += tmp32;
tmp32 = state[1];
tmp64 -= tmp32;
state[1] = (u32)tmp64;
tmp64 >>= 32; // Arithmetic shift
tmp64 += tmp32;
tmp32 = state[2];
tmp64 -= tmp32;
state[2] = (u32)tmp64;
tmp64 >>= 32; // Arithmetic shift
tmp64 += tmp32;
tmp32 = state[3];
tmp64 -= tmp32;
state[3] = (u32)tmp64;

return (state[3] ^ state[2]);
}

//------------------------------------------------
// State is [k0].....k[11]
// 
// state u32[12] = 384 bit state
// state[0..3] is 128 bit NLCG
// state[4..11] is 256 bit hash
//
// Generates u32[2] pseudorandom number 
//
// Irreversible function 256 bit strong
//
void keystream_generator(u32 *output, u32 *state)
{
const u32 rounds = 20;
u32 lock_bar[8];

//---- State modification ---

// NLCG

nlcg128(state);

// Hash function

copy(lock_bar, state + 4, 8);

XXTEA_Encode(state + 4, 8, state, rounds);

xor_data(state + 4, lock_bar, 8);

zero(lock_bar, 8); // wash previous state

// Output encoding 

output[0] = state[2];
output[1] = state[3];

IDEA_Encode(output, state + 4);
}

//===============
// Initialize keystream state u32[12] with key u32[12]
//
void keystream_generator_init(u32 *state, u32 const *key)
{
// Use 384 bits of key as is

copy(state, key, 12);

reset_stack();
}
//---------------------------------------------

KEYSTREAM_GENERATOR::KEYSTREAM_GENERATOR()
{
KEYSTREAM_GENERATOR::reset();
}

KEYSTREAM_GENERATOR::~KEYSTREAM_GENERATOR()
{
KEYSTREAM_GENERATOR::reset(); 
}

void KEYSTREAM_GENERATOR::reset()
{
cntr = 0;
skip_cntr = 0;
byte_buffer_idx = 0;
memset(byte_buffer,0, sizeof(byte_buffer));
cheap_random_number(state, 12);
reset_stack();
}

void KEYSTREAM_GENERATOR::init_key(u32 const* key)
{
keystream_generator_init(state, key);
cntr = 0;
skip_cntr = 0;
byte_buffer_idx = 0;
}

u8 KEYSTREAM_GENERATOR::get_keystream_byte()
{

if(!byte_buffer_idx) // time to generate next 8 bytes
 {
 u32 tmp_buf[2];   
 keystream_generator(tmp_buf, state);
 
 for(u32 ci = 0; ci < 4; ci++)
  {
  byte_buffer[ci]   = (u8)(tmp_buf[0] & 0xFF);
  tmp_buf[0] >>= 8;

  byte_buffer[ci+4] = (u8)(tmp_buf[1] & 0xFF);  
  tmp_buf[1] >>= 8;
  }
 }

u8 ret_value = byte_buffer[byte_buffer_idx];
byte_buffer_idx = (byte_buffer_idx + 1) & 0x07;

cntr++;


return ret_value;
}

int KEYSTREAM_GENERATOR::counter(u64 x)
{
if(x < cntr) return 0; // Can't go back in time

while(x > cntr)
 {
 skip_cntr++;
 get_keystream_byte(); // Roll forward
 }
 
return 1;
}

u64 KEYSTREAM_GENERATOR::counter()
{
return cntr;
}

u64 KEYSTREAM_GENERATOR::skip_counter()
{
return skip_cntr;
}


//==================== <END> ========================

