#ifndef NETWORK_H_INCLUDED
#define NETWORK_H_INCLUDED
//-----------------------------------------------------------------------
#include "types.h"
#include "diffie_helman.h"
#include "kademlia.h"
//-----------------------------------------------------------------------
static const u32 NETWORK_SIZE_UDP = 384;
static const u32 NETWORK_SIZE_SECURED = NETWORK_SIZE_UDP-NETWORK_ID_LENGTH*4-4;
static const u32 NETWORK_SIZE_UNSECURED = NETWORK_SIZE_SECURED-16;
static const u32 NETWORK_SIZE_USEFUL = NETWORK_SIZE_UNSECURED-4;
static const u32 NETWORK_SIZE_TEXT = 1024;
static const u32 NETWORK_SIZE_DH_MAX = 8192;
static const u32 NETWORK_PROTOCOL_VERSION = 2;
enum {
	NETWORK_PROT_KADEMLIA,
	NETWORK_PROT_CONTROL,
	NETWORK_PROT_FILE_TRANSFER
};
enum {
	USER_STATUS_OFFLINE,
	USER_STATUS_ONLINE,
	USER_STATUS_AWAY,
	USER_STATUS_MAX
};
//-----------------------------------------------------------------------
enum {
	SESSION_ENCODED,
	SESSION_REQUEST,
	SESSION_RESPONSE,
	SESSION_PING_REQ,
	SESSION_PING_RESP,
	SESSION_NAME_REQ,
	SESSION_NAME_RESP,
	SESSION_CONFIRMATION,
	SESSION_TERMINATE,
	SESSION_NOT_ESTABLISHED,
	SESSION_NOT_CONFIRMED,
	SESSION_MAX_PACK
};

#define SESSION_USER_AWAY	0x80000000
//-----------------------------------------------------------------------
#define SESSION_LAUNCHED		0x10000000
#define SESSION_PINGED			0x20000000
#define SESSION_RESPONDED		0x40000000
#define SESSION_ESTABLISHING	0x01000000
#define SESSION_FIRST_STAGE		0x80000000
#define SESSION_JUST_SETUP		0x02000000
#define SESSION_REJECTION		0x04000000
#define SESSION_REQ_COUNT_MASK	0xFF
#define SESSION_REQ_COUNT_MAX	200
//-----------------------------------------------------------------------
typedef struct {
	u32 ip;
	u16 port;
	u32 protocol;
	u32 rec_flags;	// bit flags of received packets
	u32 buff[NETWORK_SIZE_DH_MAX/4];
	u32 last_buf[NETWORK_SIZE_DH_MAX/4];
	u32 sent_buf[NETWORK_SIZE_DH_MAX/4];
	u32 conf_buf[NETWORK_SIZE_DH_MAX/4];
	u32 rec_size;
	u32 last_size;
	u32 sent_size;
	u32 conf_size;
	TX_SECURE_SESSION tx_session;
	RX_SECURE_SESSION rx_session;
} ESTABLISHMENT;
//-----------------------------------------------------------------------
class CONNECTION {
public:
	u32 flags;
	NODE nd;
	u8 userInfo[user_info_max_length_bytes];
	bool away;
	u32 userInfoSize;
	u32 pack_num_est;  // number of packets for establishment
	u32 pack_num_sec;  // number of secured packets
	u32 pack_num_drop; // number of dropped packets
	u32 pack_num_sent; // number of sent packets
	u32 timeout;
	HANDLE rx_sign;
	HANDLE tx_sign;
	HANDLE conf_sign;

	CRITICAL_SECTION est_cs;
	ESTABLISHMENT* est;
	INITIALIZED_SESSION ses;
	CONNECTION* next_connection;
	CONNECTION(const NODE* node);
	~CONNECTION();
	bool isEstablished(void) {return (flags&SESSION_LAUNCHED) == SESSION_LAUNCHED;};
};
//-----------------------------------------------------------------------
typedef struct {
	u32* buffer;
	u32 buf_len;
	NODE* nd;
} NET_PACKET;
//-----------------------------------------------------------------------
typedef struct {
	const NODE* nd;
	u32 protocol;
	const u32* buffer;
	u32 buf_len;
} PACKET_FOR_SEND;
//-----------------------------------------------------------------------
class NETWORK {
	u32 sending_buff[NETWORK_SIZE_DH_MAX/4];
	u32 sending_size;
	u32 sending_buff_1[NETWORK_SIZE_DH_MAX/4];
	u32 sending_size_1;
	TX_SECURE_SESSION cmn_tx_session;
	TX_SECURE_SESSION cmn_tx_session_1;
	CONNECTION* con;
	HWND mainWnd;
	HANDLE eStopThreads;
	HANDLE recThread;
	HANDLE tmoThread;
	SOCKET soc;
	CRITICAL_SECTION cs;
	u32 my_id[NETWORK_ID_LENGTH];
	sockaddr_in my_addr;
	bool away;


	friend DWORD WINAPI receiverProc(NETWORK* n);
	friend DWORD WINAPI timeoutProc(NETWORK* n);
	friend DWORD WINAPI rx_signature(CONNECTION* s);
	friend DWORD WINAPI tx_signature(CONNECTION* s);
	friend DWORD WINAPI conf_signature(CONNECTION* s);


	CONNECTION* find_connection(const NODE* nd, bool establish);
	void send_big_buffer(const NODE* nd, u32* buf, u32 size, u32 command, u32* pack_cnt);
	bool collect_big_buffer(u32* cur_buf, u32 size, ESTABLISHMENT* est);

public:
	NETWORK(NODE* me, HWND mWnd);
	~NETWORK();

	u32 send_packet(PACKET_FOR_SEND* pack);

	CONNECTION* get_first_node(void) {return con;};
	CONNECTION* get_next_node(CONNECTION* cur);
	void get_user_info(NODE* nd, LPTSTR info);
	void get_counters(const NODE* nd, u32* est, u32* sec, u32* drop, u32* sent, u32* missed);
	int get_user_status(const u32* id);
	void update_ip(NODE* nd);
	void reset_sessions(void);
	void set_away(bool status) {away = status;};

	bool findUserByName(u8* userInfo, u16 size, u32* id);
	bool findNode(const u32 *id, NODE *nd);
};
//-----------------------------------------------------------------------
#endif //NETWORK_H_INCLUDED
